# extracttext.py - Extract text (OCR) from images
import os
import sys
from os import listdir
from os.path import isfile, join, exists
import re
import pymysql
import settings
import subprocess
import tempfile
import pandas as pd
import matplotlib.pyplot as plt
plt.switch_backend('agg')
import seaborn as sns
import json
import csv
import errno
import shutil
from PIL import Image, ImageDraw

# maximum pixel height (used for splitting large images)
MAX_PIXEL_HEIGHT = 4000

# get color palette
color = sns.color_palette()

# load scores
scores = None
try:
    with open('scores.json') as data_file:
        scores = json.load(data_file)
except:
    print ('Could not open scores.json')
if scores == None or len(scores) == 0:
    exit ('scores not defined. Cannot continue.')

# get db
def get_db():
    db = pymysql.connect(
        host=os.environ['DB_HOST'],
        user=os.environ['DB_USERNAME'],
        password=os.environ['DB_PASSWORD'],
        db=os.environ['DB_DATABASENAME'],
        charset='utf8mb4')

    return db

# create database
def database_created():

    # Open database connection
    try:
        db = pymysql.connect(
            host=os.environ['DB_HOST'],
            user=os.environ['DB_USERNAME'],
            passwd=os.environ['DB_PASSWORD'])

        # prepare a cursor object using cursor() method
        cursor = db.cursor()

        # drop (if exists) and create database
        cursor.execute("DROP DATABASE IF EXISTS " + os.environ['DB_DATABASENAME'])
        cursor.execute("CREATE DATABASE " + os.environ['DB_DATABASENAME'])

        # disconnect from server
        db.close()

        return True

    except:

        print ('Error connecting to database. Review .env file for connection parameters.')

        return False

# create tables
def tables_created():

    # Open database connection
    try:

        # get db
        db = get_db()

        # prepare a cursor object using cursor() method
        cursor = db.cursor()

        # Create table as per requirement
        sql = """CREATE TABLE IMAGE_OCR (
           IMAGE_NAME   CHAR(200) NOT NULL UNIQUE,
           IMAGE_WIDTH INT NOT NULL DEFAULT 0,
           IMAGE_HEIGHT INT NOT NULL DEFAULT 0,
           PHRASE_COUNT INT NOT NULL DEFAULT 0,
           SCORE        INT) DEFAULT CHARSET=utf8"""
        cursor.execute(sql)
        # Create table as per requirement
        sql = """CREATE TABLE IMAGE_OCR_PHRASE (
           IMAGE_NAME           CHAR(200) NOT NULL,
           OCR_PHRASE_TYPE      CHARACTER(10) NOT NULL,
           OCR_PHRASE           TEXT,
           SCORE                INT,
           X1                   INT,
           Y1                   INT,
           X2                   INT,
           Y2                   INT,
           CONTEXT              CHAR(200) NOT NULL) DEFAULT CHARSET=utf8"""
        cursor.execute(sql)

        # disconnect from server
        db.close()

        return True

    except:

        print ('Error connecting to database. Review .env file for connection parameters and create database named in DB_DATABASENAME.')

        return False


#stitch two images vertically
def stitch_images(image_one, image_two, output):
    try:
        subprocess.check_call(["convert", image_one, image_two, "-append", output])
        return True
    except:
        return False


# split large images
def split_image(input, output, height):

    try:

        # negate (reverse black and white), noise removal (Erode), tile image
        params = ["+gravity", "-crop", "x" + str(height)]
        # params = ["-morphology", "Open", "Disk:1", "-colorspace", "gray", "-auto-level", "-threshold", "90%", "-density", "300", "+gravity", "-crop", "x" + str(height)]
        # params = ["-colorspace", "gray", "+dither", "-colors", "16", "-type", "bilevel", "-density", "300", "+gravity", "-crop", "x" + str(height)]
        # params = ["-sharpen", "2", "-depth", "8", "+gravity", "-crop", "x" + str(height)]

        subprocess.check_call(["convert", input] + params + [output])

        return True

    except:

        return False

# save image ocr into database
def save_image_ocr(image_file, phrase_list, image_size):

    # Open database connection
    try:

        # get db
        db = get_db()

        # prepare a cursor object using cursor() method
        cursor = db.cursor()

        # calculate image_score
        phrase_count = 0
        image_score = 0

        for phrase in phrase_list:
            sql = "INSERT INTO `IMAGE_OCR_PHRASE` (`IMAGE_NAME`, `OCR_PHRASE`, SCORE, OCR_PHRASE_TYPE, X1, Y1, X2, Y2, CONTEXT) VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s) "
            cursor.execute(sql, (image_file, phrase['text'], str(phrase['score']), phrase['type'], str(phrase['x1']), str(phrase['y1']), str(phrase['x2']), str(phrase['y2']), phrase['context'] ))

            phrase_count += 1
            image_score += phrase['score']

        print ('   phrase count:', phrase_count, 'for image file', image_file)

        # save score by image
        sql = """
            INSERT INTO `IMAGE_OCR` (`IMAGE_NAME`, `IMAGE_WIDTH`, `IMAGE_HEIGHT`, `SCORE`, `PHRASE_COUNT`)
            VALUES (%s, %s, %s, %s, %s)
        """
        cursor.execute(sql, (image_file, image_size[0], image_size[1], image_score, phrase_count))

        # commit changes
        db.commit()

        # disconnect from server
        db.close()

        return True

    except:
        print("Unexpected error:", sys.exc_info()[0], sys.exc_info()[1])

        return False

# extract text from a directory of images
def extract_text(input_images_dir, output_images_dir):

    if not os.path.exists(input_images_dir):
        print("The directory you entered doesn't exist: " + input_images_dir)
        return

    try:
        if not os.path.exists(output_images_dir):
            os.makedirs(output_images_dir)
        else:
            shutil.rmtree(output_images_dir)
            os.makedirs(output_images_dir)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

    image_files = [f for f in listdir(input_images_dir) if isfile(join(input_images_dir, f))]
    # text_output_dict should contain one string for each file - empty strings are expected where OCR retreival
    # was unsuccessful
    text_output_dict = {}
    for image_file in image_files:

        full_image_file = os.path.join(input_images_dir, image_file)
        full_highlight_image_file = None;

        print('Processing', full_image_file)

        identify_cmd = ["identify", full_image_file]
        identify_result = str(subprocess.check_output(identify_cmd))
        image_size = identify_result.split()[2].split('x')

        # create a temporary directory using the context manager
        with tempfile.TemporaryDirectory() as tmpdirname:
            print('created temporary directory', tmpdirname)

            # split image max pixels height
            if split_image(full_image_file,
                os.path.join(tmpdirname, image_file + "_tiles_%d.tif"), MAX_PIXEL_HEIGHT):

                # process sub image files
                sub_image_files = [f for f in listdir(tmpdirname) if isfile(join(tmpdirname, f))]
                #sorting the filenames by the integer value at the end of name
                sub_image_files.sort(key=lambda filename: int(filename[filename.rindex('_') + 1: -4]))
                ocr_text = ""
                phrase_list = []
                vertical_pixel_offset = 0
                for sub_image_file in sub_image_files:
                    # ocr image
                    full_sub_image_file = os.path.join(tmpdirname, sub_image_file)
                    print('   ocr sub image', full_sub_image_file)

                    full_sub_image_hocr_file = os.path.join(tmpdirname, sub_image_file.replace('.tif', ''))

                    """
                    Usage:
                      tesseract --help | --help-psm | --version
                      tesseract --list-langs [--tessdata-dir PATH]
                      tesseract --print-parameters [options...] [configfile...]
                      tesseract imagename|stdin outputbase|stdout [options...] [configfile...]

                    OCR options:
                      --tessdata-dir PATH   Specify the location of tessdata path.
                      --user-words PATH     Specify the location of user words file.
                      --user-patterns PATH  Specify the location of user patterns file.
                      -l LANG[+LANG]        Specify language(s) used for OCR.
                      -c VAR=VALUE          Set value for config variables.
                                            Multiple -c arguments are allowed.
                      -psm NUM              Specify page segmentation mode.
                    NOTE: These options must occur before any configfile.

                    Page segmentation modes:
                      0    Orientation and script detection (OSD) only.
                      1    Automatic page segmentation with OSD.
                      2    Automatic page segmentation, but no OSD, or OCR.
                      3    Fully automatic page segmentation, but no OSD. (Default)
                      4    Assume a single column of text of variable sizes.
                      5    Assume a single uniform block of vertically aligned text.
                      6    Assume a single uniform block of text.
                      7    Treat the image as a single text line.
                      8    Treat the image as a single word.
                      9    Treat the image as a single word in a circle.
                     10    Treat the image as a single character.

                    Single options:
                      -h, --help            Show this help message.
                      --help-psm            Show page segmentation modes.
                      -v, --version         Show version information.
                      --list-langs          List available languages for tesseract engine.
                      --print-parameters    Print tesseract parameters to stdout.

                    """

                    # generate hocr output (saved into result.hocr) from ocr process
                    # using tesseract command
                    #   --user-words use word list
                    # also define inline parameters:
                    #   language_model_penalty_non_freq_dict_word	0.2     Penalty for words not in the frequent word dictionary
                    #   language_model_penalty_non_dict_word	    0.3	    Penalty for non-dictionary words
                    tesseract_cmd = ["tesseract", full_sub_image_file, "--user-words", "user-words.txt",
                                        "-c", "language_model_penalty_non_freq_dict_word=0.2 language_model_penalty_non_dict_word=0.3 enable_new_segsearch=1",
                                        full_sub_image_hocr_file, "hocr"]
                    subprocess.check_call(tesseract_cmd)

                    hocr_data = None
                    with open(full_sub_image_hocr_file + '.hocr', encoding='utf-8') as hocr_file:
                        hocr_data = hocr_file.read()

                    # parse hocr
                    from html.parser import HTMLParser

                    class MyHTMLParser(HTMLParser):

                        def __init__(self, scores):
                            HTMLParser.__init__(self)
                            self.datas = []
                            self.scores = scores
                            self.found = 0
                            self.attrs = None
                            self.phrases = []

                        def handle_starttag(self, tag, attrs):

                            for attr in attrs:
                                if attr[0] == 'class' and attr[1] == 'ocrx_word':
                                    self.attrs = attrs

                        def handle_data(self, data):

                            data = data.replace('\n', '').replace(',', ' ').replace(';', ' ').replace('.', ' ').replace('0', 'o').replace('im', 'in').replace('om', 'on').replace('mo', 'no').replace('wm', 'wn').replace('sn', 'sh').strip().lower()
                            if len(data) > 0:
                                for attr in self.attrs:
                                    if attr[0] == 'title':
                                        ts = attr[1].replace(';', ' ').split(' ')
                                        x1 = int(ts[1])
                                        y1 = int(ts[2])
                                        x2 = int(ts[3])
                                        y2 = int(ts[4])

                                new_phrases = data.split()
                                for new_phrase in new_phrases:
                                    if new_phrase:
                                        self.datas.append({ 'text': new_phrase, 'x1': x1, 'y1': y1, 'x2': x2, 'y2': y2 })

                    source_img = Image.open(full_sub_image_file).convert("RGBA")

                    # if there is hocr data...
                    if hocr_data != None:

                        # parse hocr (ocr info)
                        parser = MyHTMLParser(scores)
                        parser.feed(hocr_data)

                        max_num_words = 1
                        for s in parser.scores:
                            for p in s['phrases']:
                                max_num_words = max(max_num_words, len(p.split()))

                        last_n_phrases = [''] * max_num_words

                        for idx in range(0, len(parser.datas)):
                            added = 0

                            for i in reversed(range(1, len(last_n_phrases))):
                                last_n_phrases[i] = last_n_phrases[i - 1] + ' ' + parser.datas[idx]['text']
                            last_n_phrases[0] = parser.datas[idx]['text']

                            for i in reversed(range(0, min(len(last_n_phrases), idx))):
                                for s in parser.scores:
                                    for p in s['phrases']:
                                        if not added:
                                            if last_n_phrases[i] == p:
                                                phrase = {}
                                                phrase['text'] = p
                                                phrase['type'] = s['type']
                                                phrase['color'] = s['color']
                                                phrase['score'] = s['score']
                                                phrase['x1'] = min([x['x1'] for x in parser.datas[idx-i:idx+1]])
                                                phrase['y1'] = min([x['y1'] for x in parser.datas[idx-i:idx+1]])
                                                phrase['x2'] = max([x['x2'] for x in parser.datas[idx-i:idx+1]])
                                                phrase['y2'] = max([x['y2'] for x in parser.datas[idx-i:idx+1]])
                                                phrase['context'] = last_n_phrases[-1]
                                                parser.phrases.append(phrase)
                                                added = 1

                        # draw ocr info on sub-image
                        draw = ImageDraw.Draw(source_img, mode="RGBA")
                        for phrase in parser.phrases:
                            boxpts = (
                                (phrase['x1'], phrase['y1']),
                                (phrase['x2'], phrase['y1']),
                                (phrase['x2'], phrase['y2']),
                                (phrase['x1'], phrase['y2']),
                                (phrase['x1'], phrase['y1']),
                            )
                            draw.line(boxpts, fill=phrase['color'], width=3)

                            phrase['y1'] += vertical_pixel_offset
                            phrase['y2'] += vertical_pixel_offset

                        phrase_list.extend(parser.phrases)

                    # save highlight output jpg image
                    full_highlight_sub_image_file = os.path.join(tmpdirname, sub_image_file.replace('.tif', '') + '.png')
                    source_img.save(full_highlight_sub_image_file, "PNG")

                    if full_highlight_image_file is None:
                        full_highlight_image_file = os.path.join(output_images_dir, image_file.replace('.tif', '') + '.png')
                        source_img.save(full_highlight_image_file, "PNG")
                        print ('Highlighted OCR results saved to', full_highlight_image_file)
                    else:
                        stitch_images(full_highlight_image_file, full_highlight_sub_image_file, full_highlight_image_file)
                        print ('Highlighted OCR results stitched to', full_highlight_image_file)

                    vertical_pixel_offset += MAX_PIXEL_HEIGHT


                # save results
                save_image_ocr(image_file, phrase_list, image_size)
                text_output_dict[image_file] = ocr_text

            else:
                print ('Error processing image')

                # save results
                save_image_ocr(image_file, [], image_size)
                text_output_dict[image_file] = None

    return text_output_dict

# report by image
def image_report(image_file):

    # Open database connection
    try:

        # get db
        db = get_db()

        # select phrase score
        with db.cursor() as cursor:
            sql = """
                SELECT `IMAGE_NAME`, `OCR_PHRASE`, SUM(SCORE), OCR_PHRASE_TYPE, COUNT(OCR_PHRASE)
                  FROM `IMAGE_OCR_PHRASE`
                 WHERE `IMAGE_NAME` = %s
                 GROUP BY `OCR_PHRASE`, `IMAGE_NAME`, `SCORE`, `OCR_PHRASE_TYPE`"""
            cursor.execute(sql, (image_file,))
            result = cursor.fetchall()
            for r in result:
                print ('{:12d}'.format(int(r[4])), ' ', '{:6d}'.format(int(r[2])), '  <<', r[1], '>>', '(', r[3], ')')

        # disconnect from server
        db.close()

        return True

    except:
        print("Unexpected error:", sys.exc_info()[0], sys.exc_info()[1])

        return False


# print summary report
def summary_report():

    # Open database connection
    try:

        # get db
        db = get_db()

        # select image score
        with db.cursor() as cursor:
            sql = "SELECT IMAGE_NAME, SCORE, PHRASE_COUNT FROM `IMAGE_OCR` ORDER BY SCORE DESC"
            cursor.execute(sql)
            result = cursor.fetchall()

            # print image score
            if (len(result) > 0):
                for r in result:
                    print ('=====================================')
                    print ('Image name:', r[0])
                    print ('Phrase count    Score   Phrase (type)')
                    print ('------------    -----   ------------------------------')
                    image_report(r[0])
                    print ('Total Score for image name:', r[0])
                    print ('Phrase count    Score   ')
                    print ('------------    -----   ')
                    print ('{:12d}'.format(r[2]), ' ', '{:6d}'.format(r[1]))
                    print ('')
            else:
                print ('No data to report')

        # disconnect from server
        db.close()

        return True

    except:
        print("Unexpected error:", sys.exc_info()[0], sys.exc_info()[1])

        return False

# generate summary chart
def summary_chart(output_images_dir):

    # Open database connection
    try:

        # get db
        db = get_db()

        # query data
        df = pd.read_sql('SELECT IMAGE_NAME, SCORE FROM `IMAGE_OCR` WHERE SCORE > ' +
                         os.environ['SCORE_THERESHOLD'] + ' ORDER BY SCORE DESC', con=db)

        # disconnect from server
        db.close()

        # if there's data to plot
        if (len(df) > 0):
            plt.switch_backend('agg')
            plt.figure()
            sns.barplot(y="IMAGE_NAME", x="SCORE", data=df, color=color[4])
            plt.ylabel('Image', fontsize=12)
            plt.xlabel('Score', fontsize=18)
            plt.title("Image by score", fontsize=18)
            plt.tight_layout()
            full_chart_image_file = os.path.join(output_images_dir, 'chart-report.png')
            plt.savefig(full_chart_image_file)
            print ('Chart saved to', full_chart_image_file)
        else:
            print ('No data to plot')

        return True

    except:
        print("Unexpected error:", sys.exc_info()[0], sys.exc_info()[1])

        return False

# if 'report' is passed as first argument
if sys.argv[1] == 'report':

    # reports
    print ('reports...')

    # create summary report
    summary_report()

    # create summary chart
    output_images_dir = sys.argv[2]
    summary_chart(output_images_dir)
else:

    # launch OCR process
    if database_created() and tables_created():

        # process
        print ('process...')

        # call extract text, pass input image dir, output image dir
        # and save returned dictionary (future versions will use this)
        input_images_dir = sys.argv[1]
        output_images_dir = sys.argv[2]
        text_output_dict = extract_text(input_images_dir, output_images_dir)

        # create summary report
        summary_report()

        # create summary chart
        summary_chart(output_images_dir)
